#ifndef H
#define H

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <delay.h>

struct Packet {
    uint16_t start_code;
    uint8_t address[4];
    uint8_t type;
    uint16_t length;
    uint8_t data[20];
};


#define FINGERPRINT_OK 0x00
#define FINGERPRINT_PACKETRECIEVEERR 0x01
#define FINGERPRINT_NOFINGER 0x02
#define FINGERPRINT_IMAGEFAIL 0x03
#define FINGERPRINT_IMAGEMESS 0x06
#define FINGERPRINT_FEATUREFAIL 0x07
#define FINGERPRINT_NOMATCH 0x08
#define FINGERPRINT_NOTFOUND 0x09
#define FINGERPRINT_ENROLLMISMATCH 0x0A
#define FINGERPRINT_BADLOCATION 0x0B
#define FINGERPRINT_DBRANGEFAIL 0x0C
#define FINGERPRINT_UPLOADFEATUREFAIL 0x0D
#define FINGERPRINT_PACKETRESPONSEFAIL 0x0E
#define FINGERPRINT_UPLOADFAIL 0x0F
#define FINGERPRINT_DELETEFAIL 0x10
#define FINGERPRINT_DBCLEARFAIL 0x11
#define FINGERPRINT_PASSFAIL 0x13
#define FINGERPRINT_INVALIDIMAGE 0x15
#define FINGERPRINT_FLASHERR 0x18
#define FINGERPRINT_INVALIDREG 0x1A
#define FINGERPRINT_ADDRCODE 0x20
#define FINGERPRINT_PASSVERIFY 0x21

#define FINGERPRINT_STARTCODE 0xEF01

#define FINGERPRINT_COMMANDPACKET 0x1
#define FINGERPRINT_DATAPACKET 0x2
#define FINGERPRINT_ACKPACKET 0x7
#define FINGERPRINT_ENDDATAPACKET 0x8

#define FINGERPRINT_TIMEOUT 0xFF
#define FINGERPRINT_BADPACKET 0xFE

#define FINGERPRINT_GETIMAGE 0x01
#define FINGERPRINT_IMAGE2TZ 0x02
#define FINGERPRINT_REGMODEL 0x05
#define FINGERPRINT_STORE 0x06
#define FINGERPRINT_LOAD 0x07
#define FINGERPRINT_UPLOAD 0x08
#define FINGERPRINT_DELETE 0x0C
#define FINGERPRINT_EMPTY 0x0D
#define FINGERPRINT_SETPASSWORD 0x12
#define FINGERPRINT_VERIFYPASSWORD 0x13
#define FINGERPRINT_HISPEEDSEARCH 0x1B
#define FINGERPRINT_TEMPLATECOUNT 0x1D

#define DEFAULTTIMEOUT 2000

uint8_t getImage(void);
uint8_t image2Tz(uint8_t slot);
uint8_t createModel(void);

uint8_t emptyDatabase(void);
uint8_t storeModel(uint16_t location);
uint8_t loadModel(uint16_t location);
uint8_t getModel(void);
uint8_t deleteModel(uint16_t location);
uint8_t fingerFastSearch(void);
uint8_t getTemplateCount(void);
uint8_t setPassword(uint32_t password);
uint8_t getStructuredPacket(struct Packet *respon, uint16_t timeout);

uint16_t fingerID;
uint16_t confidence;
uint16_t templateCount;

uint8_t checkPassword(void);
//uint32_t thePassword;
//uint32_t theAddress;
//uint8_t recvPacket[20];

uint8_t *rxCounter;
uint8_t *rxData;



struct Packet packet;

void fingerPrintBegin(uint8_t *SerialData, uint8_t *serialCounter)
{
    packet.start_code = FINGERPRINT_STARTCODE;
    packet.address[0] = 0xFF; 
    packet.address[1] = 0xFF;
    packet.address[2] = 0xFF; 
    packet.address[3] = 0xFF;

    rxData = SerialData;
    rxCounter = serialCounter;
}


void uart_transmit (unsigned char data)
{
    while (!( UCSRA & (1<<UDRE)));                // wait while register is free
    UDR = data;                                   // load data in the register
}

void kirimPacket()
{
    uint16_t crc;     
    uint16_t len = packet.length + 2;   
    uint8_t i;  

    uart_transmit (packet.start_code >> 8);
    uart_transmit (packet.start_code & 0xFF);
    uart_transmit (packet.address[0]);
    uart_transmit (packet.address[1]);
    uart_transmit (packet.address[2]);
    uart_transmit (packet.address[3]);
    uart_transmit (packet.type);
    uart_transmit (len >> 8);
    uart_transmit (len & 0xFF);
    

    crc = packet.type + (len>>8) + (len&0xFF);
    for (i=0; i< packet.length; i++) {
        uart_transmit(packet.data[i]);
        crc += packet.data[i];
    }

    uart_transmit (crc >> 8);
    uart_transmit (crc & 0xFF);   
    
    *rxCounter = 0;      
} 


uint8_t checkPassword(void) 
{
  struct Packet respon;   

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 5;   
  packet.data[0] = FINGERPRINT_VERIFYPASSWORD;
  packet.data[1] = 0;
  packet.data[2] = 0;
  packet.data[3] = 0;
  packet.data[4] = 0;
  
  kirimPacket(); 

  if(getStructuredPacket(&respon, DEFAULTTIMEOUT) == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {
    return 0xFF;
  }
}

uint8_t getImage(void) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 1;
  packet.data[0] = FINGERPRINT_GETIMAGE;
  

  kirimPacket();  
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t image2Tz(uint8_t slot) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 2;
  packet.data[0] = FINGERPRINT_IMAGE2TZ;
  packet.data[1] = slot;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t createModel(void) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 1;
  packet.data[0] = FINGERPRINT_REGMODEL;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t storeModel(uint16_t location) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 4;
  packet.data[0] = FINGERPRINT_STORE;
  packet.data[1] = 1;
  packet.data[2] = location >> 8;
  packet.data[3] = location & 0xFF;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t loadModel(uint16_t location) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 4;
  packet.data[0] = FINGERPRINT_LOAD;
  packet.data[1] = 1;
  packet.data[2] = location >> 8;
  packet.data[3] = location & 0xFF;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t getModel(void) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 2;
  packet.data[0] = FINGERPRINT_UPLOAD;
  packet.data[1] = 1;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t deleteModel(uint16_t location) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 5;
  packet.data[0] = FINGERPRINT_DELETE;
  packet.data[1] = location >> 8;
  packet.data[2] = location & 0xFF;
  packet.data[3] = 0;
  packet.data[4] = 1;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t emptyDatabase(void) {
  struct Packet respon;
  uint8_t ret;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 1;
  packet.data[0] = FINGERPRINT_EMPTY;

  kirimPacket();
  ret = getStructuredPacket(&respon, DEFAULTTIMEOUT); 
  if(ret == FINGERPRINT_OK)
  {
    return respon.data[0];
  }               
  else
  {              
    return ret;
  }
}

uint8_t fingerFastSearch(void) {
  struct Packet respon;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 6;
  packet.data[0] = FINGERPRINT_HISPEEDSEARCH;
  packet.data[1] = 1;
  packet.data[2] = 0;
  packet.data[3] = 0;
  packet.data[4] = 0;
  packet.data[5] = 0xA3;

  kirimPacket();
  getStructuredPacket(&respon, DEFAULTTIMEOUT);

  fingerID = 0xFFFF;
  confidence = 0xFFFF;

  fingerID = respon.data[1];
  fingerID <<= 8;
  fingerID |= respon.data[2];

  confidence = respon.data[3];
  confidence <<= 8;
  confidence |= respon.data[4];

  return respon.data[0];
}

uint8_t getTemplateCount(void) {
  struct Packet respon;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 1;
  packet.data[0] = FINGERPRINT_TEMPLATECOUNT;

  kirimPacket();
  getStructuredPacket(&respon, DEFAULTTIMEOUT);

  templateCount = respon.data[1];
  templateCount <<= 8;
  templateCount |= respon.data[2];

  return respon.data[0];
}
uint8_t setPassword(uint32_t password) {
  struct Packet respon;

  packet.type = FINGERPRINT_COMMANDPACKET;
  packet.length = 5;
  packet.data[0] = FINGERPRINT_SETPASSWORD;
  packet.data[1] = password >> 24;
  packet.data[2] = password >> 16;
  packet.data[3] = password >> 8;
  packet.data[4] = password &0xFF;

  kirimPacket();
  return getStructuredPacket(&respon, DEFAULTTIMEOUT);
}

uint8_t getStructuredPacket(struct Packet *respon, uint16_t timeout) {
    uint8_t byte;
    uint16_t idx=0, timer=0;
    uint8_t readCounter = 0;   
    
    while(1) 
    {
        while(*rxCounter == readCounter) 
        {
            delay_ms(1);
            timer++; 
            if( timer >= timeout) 
            {
                return FINGERPRINT_TIMEOUT;
            }
        }
        byte = rxData[readCounter];
        readCounter++;
        

        switch (idx) 
        {
            case 0:
                if (byte != (FINGERPRINT_STARTCODE >> 8))
                    continue;
                respon->start_code = (uint16_t)byte << 8;
                break;
            case 1:
                respon->start_code |= byte;
                if (respon->start_code != FINGERPRINT_STARTCODE)
                { 
                    return FINGERPRINT_BADPACKET;
                }
                break;
            case 2:
            case 3:
            case 4:
            case 5:
                respon->address[idx-2] = byte;
                break;
            case 6: 
                respon->type = byte; 
                break;
            case 7: 
                respon->length = (uint16_t)byte << 8; 
                break;
            case 8: 
                respon->length |= byte; 
                break;
            default:  
            
                respon->data[idx-9] = byte;  
                if((idx-8) == respon->length)
                    return FINGERPRINT_OK;
                break;
        }
        idx++;
    }
    return 0x81;
}
  

#endif
